from sqlalchemy import create_engine, event
from pydantic import validator

from . import Database

db_parameters = ["database"]


class Sqlite(Database):
    def feature(self, feature):
        return feature in (
            "CANNOT ALTER INDEXES",
            "CANNOT SET SCHEMA",
            "CANNOT SPECIFY DDL IN SELECT",
            "NO SCHEMA SUPPORT",
        )

    def create_engine(self, settings):
        database = settings.pop("database")

        engine = create_engine(f"sqlite:///{database}", **settings)

        # this is set to fix a SQLite setting which can prevent a second execution of SAYN.
        # More info on this command here: https://sqlite.org/pragma.html#pragma_legacy_alter_table
        @event.listens_for(engine, "connect")
        def do_connect(dbapi_connection, connection_record):
            cursor = dbapi_connection.cursor()
            cursor.execute("PRAGMA legacy_alter_table = ON")
            cursor.close()

        return engine

    def execute(self, script):
        with self.engine.connect().execution_options(autocommit=True) as connection:
            connection.connection.executescript(script)



from pathlib import Path

from pydantic import BaseModel, FilePath, validator
from typing import Optional

from ..core.errors import Ok, Err, Exc
from ..database import Database
from . import Task


class Config(BaseModel):
    sql_folder: Path
    file_name: FilePath
    db: Optional[str]

    @validator("file_name", pre=True)
    def file_name_plus_folder(cls, v, values):
        return Path(values["sql_folder"], v)


class SqlTask(Task):
    @property
    def target_db(self):
        return self.connections[self._target_db]

    def use_db_object(
        self, name, schema=None, tmp_schema=None, db=None, request_tmp=True
    ):
        if db is None:
            target_db = self.target_db
        elif isinstance(db, str):
            target_db = self.connections[db]
        elif isinstance(db, Database):
            target_db = db
        else:
            return Err("use_db_object", "wrong_connection_type")

        target_db._request_object(
            name,
            schema=schema,
            tmp_schema=tmp_schema,
            task_name=self.name,
            request_tmp=request_tmp,
        )

    def setup(self, **config):
        conn_names_list = [
            n for n, c in self.connections.items() if isinstance(c, Database)
        ]

        # set the target db for execution
        if config.get("db") is not None:
            if config["db"] not in conn_names_list:
                return Err("task_definition", "db_not_in_settings", db=config["db"])
            self._target_db = config["db"]
        else:
            self._target_db = self._default_db

        self.config = Config(sql_folder=self.run_arguments["folders"]["sql"], **config)

        result = self.compile_obj(self.config.file_name)
        if result.is_err:
            return result
        else:
            self.sql_query = result.value

        if self.run_arguments["command"] == "run":
            self.set_run_steps(["Write Query", "Execute Query"])
        else:
            self.set_run_steps(["Write Query"])

        return Ok()

    def compile(self):
        with self.step("Write Query"):
            result = self.write_compilation_output(self.sql_query)
            if result.is_err:
                return result

        return Ok()

    def run(self):
        with self.step("Write Query"):
            result = self.write_compilation_output(self.sql_query)
            if result.is_err:
                return result

        with self.step("Execute Query"):
            try:
                self.target_db.execute(self.sql_query)
            except Exception as e:
                return Exc(e)

        return Ok()



from copy import deepcopy
import csv
import datetime
import decimal
import io
from typing import List, Optional

from pydantic import validator
from sqlalchemy import create_engine
from sqlalchemy.sql import sqltypes

from . import Database, DDL

db_parameters = ["project", "credentials_path", "location", "dataset"]


class BigqueryDDL(DDL):
    partition: Optional[str]
    cluster: Optional[List[str]]

    @validator("cluster")
    def validate_cluster(cls, v, values):
        if len(values.get("columns")) > 0:
            missing_columns = set(v) - set([c.name for c in values.get("columns")])
            if len(missing_columns) > 0:
                raise ValueError(
                    f'Cluster contains columns not specified in the ddl: "{missing_columns}"'
                )

        return v

    def get_ddl(self):
        return {
            "columns": [c.dict() for c in self.columns],
            "indexes": {},
            "permissions": self.permissions,
            "partition": self.partition,
            "cluster": self.cluster,
            "primary_key": list(),
        }


class Bigquery(Database):
    ddl_validation_class = BigqueryDDL

    project = None
    dataset = None

    def feature(self, feature):
        return feature in (
            "CAN REPLACE TABLE",
            "CAN REPLACE VIEW",
            "CANNOT CHANGE SCHEMA",
        )

    def create_engine(self, settings):
        settings = deepcopy(settings)
        self.project = settings.pop("project")

        url = f"bigquery://{self.project}"
        if "dataset" in settings:
            self.dataset = settings.pop("dataset")
            url += "/" + self.dataset

        return create_engine(url, **settings)

    def _py2sqa(self, from_type):
        python_types = {
            int: sqltypes.Integer,
            str: sqltypes.String,
            float: sqltypes.Float,
            decimal.Decimal: sqltypes.Numeric,
            datetime.datetime: sqltypes.TIMESTAMP,
            bytes: sqltypes.LargeBinary,
            bool: sqltypes.Boolean,
            datetime.date: sqltypes.Date,
            datetime.time: sqltypes.Time,
            datetime.timedelta: sqltypes.Interval,
            list: sqltypes.ARRAY,
            dict: sqltypes.JSON,
        }

        if from_type not in python_types:
            raise ValueError(f'Type not supported "{from_type}"')
        else:
            return python_types[from_type]().compile(dialect=self.engine.dialect)

    def _load_data_batch(self, table, data, schema):
        full_table_name = (
            f"{self.project}.{self.dataset if schema is None else schema}.{table}"
        )

        buffer = io.StringIO()
        writer = csv.DictWriter(buffer, fieldnames=data[0].keys())
        writer.writeheader()
        writer.writerows(data)
        buffer = io.BytesIO(buffer.getvalue().encode("utf-8"))

        from google.cloud import bigquery

        job_config = bigquery.LoadJobConfig(
            source_format=bigquery.SourceFormat.CSV,
            skip_leading_rows=1,
            # autodect=True,
        )

        client = self.engine.raw_connection()._client
        job = client.load_table_from_file(
            buffer, full_table_name, job_config=job_config
        )
        job.result()

    def move_table(self, src_table, dst_table, src_schema=None, dst_schema=None, **ddl):
        full_src_table = (
            f"{src_schema + '.' if src_schema is not None else ''}{src_table}"
        )
        select = f"SELECT * FROM {full_src_table}"
        create_or_replace = self.create_table(
            dst_table, dst_schema, select=select, replace=True, **ddl
        )

        return "\n\n".join((create_or_replace, f"DROP TABLE {full_src_table}"))



from typing import Any, Dict

from pydantic import ValidationError
from ruamel.yaml.error import MarkedYAMLError
import sqlalchemy


class Error:
    kind: str
    code: str
    details: Dict = dict()

    def __init__(self, kind, code, details):
        self.kind = kind
        self.code = code
        self.details = details

    def __repr__(self):
        return f"Result.Err ({self.kind}::{self.code}): {self.details.__repr__()}"


class Result:
    is_ok: bool = False
    value: Any = None
    error: Error = None

    def __init__(self, value: Any = None, error: Error = None):
        if error is not None:
            self.is_ok = False
            self.error = error
        else:
            self.is_ok = True
            self.value = value

    def __repr__(self):
        if self.is_ok:
            return f"Result.Ok: {self.value.__repr__()}"
        else:
            return self.error.__repr__()

    @property
    def is_err(self):
        return not self.is_ok

    def mk_error_message(self, errors):
        """Returns a list of error messages from a Error result

        Args:
          errors (List[Result]): a list of results
        """
        pass


def Ok(value=None):
    """Creates an OK result"""
    return Result(value=value)


def Err(kind, code, **details):
    """Creates an Error result"""
    return Result(error=Error(kind, code, details))


def Exc(exc, **kwargs):
    """Creates an Error result from an exception"""
    if isinstance(exc, MarkedYAMLError):
        # Ruamel error
        # TODO check that it's controlled from sayn.core.config
        return Result(
            error=Error(
                "parsing",
                "yaml_parse",
                dict(
                    **kwargs,
                    **{"error": exc.problem, "line": exc.problem_mark.line + 1},
                ),
            )
        )

    elif isinstance(exc, ValidationError):
        # Pydantic error
        return Result(
            error=Error("parsing", "validation_error", {"errors": exc.errors()})
        )

    elif isinstance(exc, NotImplementedError) and exc.args[0] == "SAYN task":
        # Missing implementation for SAYN command
        return Result(
            error=Error(
                "exception",
                "not_implemented",
                {"class": exc.args[1], "method": exc.args[2]},
            )
        )
    elif isinstance(exc, NotImplementedError):
        # Other missing method
        return Result(
            error=Error("exception", "unknown_not_implemented", {"exception": exc})
        )

    # TODO add other exceptions like:
    #   - DB errors
    #   - Jinja
    elif isinstance(exc, sqlalchemy.exc.OperationalError):
        return Result(
            error=Error(
                "database",
                "operational_error",
                {"exception": exc, "message": " ".join(exc.args)},
            )
        )

    else:
        return Result(
            error=Error("exception", "unhandled_exception", {"exception": exc})
        )


class DBError(Exception):
    db_name = None
    db_type = None

    def __init__(self, db_name, db_type, *args, **kwargs):
        self.db_name = db_name
        self.db_type = db_type

        super(sqlalchemy.SQLAlchemyError, self).__init__(*args, **kwargs)



from pathlib import Path
from datetime import date, timedelta
import sys

import click

from .utils.python_loader import PythonLoader
from .utils.task_query import get_query
from .utils.graphviz import plot_dag
from .logging import ConsoleLogger, FancyLogger, FileLogger
from .scaffolding.init_project import sayn_init
from .core.app import App
from .core.config import (
    cleanup_compilation,
    read_project,
    read_groups,
    read_settings,
    get_tasks_dict,
)
from .core.errors import Err, Result

yesterday = date.today() - timedelta(days=1)


class CliApp(App):
    def __init__(
        self,
        command,
        debug=False,
        include=list(),
        exclude=list(),
        profile=None,
        full_load=False,
        start_dt=yesterday,
        end_dt=yesterday,
    ):
        # STARTING APP: register loggers and set cli arguments in the App object
        self.run_arguments["command"] = command

        loggers = [
            ConsoleLogger(True) if debug else FancyLogger(),
            FileLogger(
                self.run_arguments["folders"]["logs"],
                format=f"{self.run_id}|" + "%(asctime)s|%(levelname)s|%(message)s",
            ),
        ]

        self.start_app(
            loggers,
            debug=debug,
            full_load=full_load,
            start_dt=start_dt,
            end_dt=end_dt,
            profile=profile,
        )

        cleanup_compilation(self.run_arguments["folders"]["compile"])

        # SETUP THE APP: read project config and settings, interpret cli arguments and setup the dag
        self.tracker.start_stage("setup")

        # Read the project configuration
        project = self.check_abort(read_project())
        groups = self.check_abort(read_groups(project.groups))
        self.set_project(project)
        settings = self.check_abort(read_settings())
        self.check_abort(self.set_settings(settings))

        # Set python environment
        self.python_loader = PythonLoader()
        if Path(self.run_arguments["folders"]["python"]).is_dir():
            self.check_abort(
                self.python_loader.register_module(
                    "python_tasks", self.run_arguments["folders"]["python"]
                )
            )

        # Set tasks and dag from it
        tasks_dict = self.check_abort(get_tasks_dict(project.presets, groups))
        task_query = self.check_abort(
            get_query(tasks_dict, include=include, exclude=exclude)
        )
        self.check_abort(self.set_tasks(tasks_dict, task_query))

        self.tracker.finish_current_stage(
            tasks={k: v.status for k, v in self.tasks.items()}
        )

    def check_abort(self, result):
        """Interpret the result of setup opreations returning the value if `result.is_ok`.

        Setup errors from the cli result in execution abort.

        Args:
          result (sayn.errors.Result): The result of a setup operation
        """
        if result is None or not isinstance(result, Result):
            self.finish_app(error=Err("app_setup", "unhandled_error", result=result))
            sys.exit()
        elif result.is_err:
            self.finish_app(result)
            sys.exit()
        else:
            return result.value


# Click arguments

click_debug = click.option(
    "--debug", "-d", is_flag=True, default=False, help="Include debug messages"
)


def click_filter(func):
    func = click.option(
        "--tasks",
        "-t",
        multiple=True,
        help="Task query to INCLUDE in the execution: [+]task_name[+], group:group_name, tag:tag_name",
    )(func)
    func = click.option(
        "--exclude",
        "-x",
        multiple=True,
        help="Task query to EXCLUDE in the execution: [+]task_name[+], group:group_name, tag:tag_name",
    )(func)
    return func


def click_incremental(func):
    func = click.option(
        "--full-load", "-f", is_flag=True, default=False, help="Do a full load"
    )(func)
    func = click.option(
        "--start-dt",
        "-s",
        type=click.DateTime(formats=["%Y-%m-%d"]),
        default=str(yesterday),
        help="For incremental loads, the start date",
    )(func)
    func = click.option(
        "--end-dt",
        "-e",
        type=click.DateTime(formats=["%Y-%m-%d"]),
        default=str(yesterday),
        help="For incremental loads, the end date",
    )(func)
    return func


def click_run_options(func):
    func = click_debug(func)
    func = click.option("--profile", "-p", help="Profile from settings to use")(func)
    func = click_incremental(func)
    func = click_filter(func)
    return func


@click.group(help="SAYN management tool.")
def cli():
    pass


# Click commands


@cli.command(help="Initialise a SAYN project in working directory.")
@click.argument("sayn_project_name")
def init(sayn_project_name):
    sayn_init(sayn_project_name)


@cli.command(help="Compile sql tasks.")
@click_run_options
def compile(debug, tasks, exclude, profile, full_load, start_dt, end_dt):
    app = CliApp("compile", debug, tasks, exclude, profile, full_load, start_dt, end_dt)
    app.compile()


@cli.command(help="Run SAYN tasks.")
@click_run_options
def run(debug, tasks, exclude, profile, full_load, start_dt, end_dt):
    app = CliApp("run", debug, tasks, exclude, profile, full_load, start_dt, end_dt)
    app.run()


@cli.command(help="Generate DAG image.")
@click_debug
@click_filter
def dag_image(debug, tasks, exclude):
    def handle_error():
        print("Errors detected in project. Run `sayn compile` to see the errors")
        sys.exit()

    result = read_project()
    if result.is_err:
        handle_error()
    else:
        project = result.value

    result = read_groups(project.groups)
    if result.is_err:
        handle_error()
    else:
        groups = result.value

    result = get_tasks_dict(project.presets, groups)
    if result.is_err:
        handle_error()
    else:
        tasks_dict = result.value
    dag = {
        task["name"]: [p for p in task.get("parents", list())]
        for task in tasks_dict.values()
    }

    plot_dag(dag, "images", "dag")

    print("Dag image created in `images/dag.png`")



from copy import deepcopy
import json
import os
from pathlib import Path
import re
import shutil
from typing import List, Optional, Dict, Any

from pydantic import BaseModel, validator, ValidationError
from ruamel.yaml import YAML
from ruamel.yaml.error import MarkedYAMLError

from ..database.creator import create as create_db
from ..utils.misc import merge_dicts, merge_dict_list
from ..utils.dag import upstream, topological_sort
from .errors import Err, Exc, Ok

RE_ENV_VAR_NAME = re.compile(r"SAYN_(?P<type>PARAMETER|CREDENTIAL)_(?P<name>.*)")


def cleanup_compilation(folder):
    compile_path = Path(folder)
    if compile_path.exists():
        if compile_path.is_dir():
            shutil.rmtree(compile_path.absolute())
        else:
            compile_path.unlink()

    compile_path.mkdir()


def read_yaml_file(filename):
    try:
        contents = Path(filename).read_text()
    except:
        return Err("parsing", "read_file_error", filename=filename)

    try:
        return Ok(YAML().load(contents))
    except MarkedYAMLError as e:
        return Exc(e, filename=filename)


def is_unique(field_name, v):
    if len(set(v)) != len(v):
        raise ValueError(f"Duplicate values found in {field_name}")
    return v


class Project(BaseModel):
    required_credentials: List[str]
    default_db: Optional[str]
    parameters: Optional[Dict[str, Any]] = dict()
    presets: Optional[Dict[str, Dict[str, Any]]] = dict()
    groups: List[str] = []

    @validator("required_credentials")
    def required_credentials_are_unique(cls, v):
        return is_unique("required_credentials", v)

    @validator("default_db", always=True)
    def default_db_exists(cls, v, values, **kwargs):
        if v is None and len(values["required_credentials"]) == 1:
            return values["required_credentials"][0]
        elif v is None:
            raise ValueError("Missing default_db in project.yaml")
        elif v not in values["required_credentials"]:
            raise ValueError(f'default_db value "{v}" not in required_credentials')

        return v

    @validator("groups", pre=True, always=True)
    def set_groups(cls, v):
        if not os.path.isdir(Path("tasks")):
            raise ValueError("'tasks' folder cannot be found in project directory.")

        groups = [f.name[:-5] for f in Path("tasks").glob("*.yaml")]

        if len(groups) == 0:
            raise ValueError(
                "No YAML file found in the tasks folder. Make sure you are using .yaml extension."
            )

        return groups


def read_project():
    result = read_yaml_file(Path("project.yaml"))
    if result.is_err:
        return result

    try:
        return Ok(Project(**result.value))
    except ValidationError as e:
        return Exc(e, where="read_project")


class TaskGroup(BaseModel):
    presets: Optional[Dict[str, Dict[str, Any]]] = dict()
    tasks: Optional[Dict[str, Dict[str, Any]]]


def read_groups(groups):
    out = dict()
    for name in groups:
        result = read_yaml_file(Path("tasks", f"{name}.yaml"))
        if result.is_err:
            return result
        else:
            try:
                out[name] = TaskGroup(**result.value)
            except ValidationError as e:
                return Exc(e, where="read_tasks")

    return Ok(out)


class Settings(BaseModel):
    class Environment(BaseModel):
        parameters: Optional[Dict[str, Any]]
        credentials: Optional[Dict[str, Dict[str, Any]]]

    class SettingsYaml(BaseModel):
        class Profile(BaseModel):
            parameters: Optional[Dict[str, Any]]
            credentials: Dict[str, str]

        credentials: Dict[str, dict]
        profiles: Dict[str, Profile]
        default_profile: Optional[str]

        @validator("profiles")
        def yaml_credentials(cls, v, values, **kwargs):
            if v is None:
                raise ValueError("No profiles defined in settings.yaml.")
            for profile_name, profile in v.items():
                for project_name, settings_name in profile.credentials.items():
                    if "credentials" not in values:
                        raise ValueError("No credentials defined in settings.yaml.")

                    if settings_name not in values["credentials"]:
                        raise ValueError(
                            f'"{settings_name}" in profile "{profile_name}" not declared in credentials.'
                        )

            return v

        @validator("default_profile", always=True)
        def default_profile_exists(cls, v, values, **kwargs):
            if "profiles" not in values:
                # We're running this always so that we can default to the first profile
                # so we cover the case of no profile specified, but the error will be
                # covered by the profiles validation
                return v

            if v is None and len(values["profiles"]) > 1:
                raise ValueError(
                    'Can\'t determine default profile. Use "default_profile" to specify it.'
                )
            elif v is None:
                return list(values["profiles"].keys())[0]

            if "profiles" not in values:
                raise ValueError("No profiles defined in settings.yaml.")
            elif v not in values["profiles"].keys():
                raise ValueError(f'default_profile "{v}" not in the profiles map.')

            return v

        def get_profile_info(self, profile_name=None):
            profile_name = profile_name or self.default_profile

            if profile_name not in self.profiles:
                raise ValueError(f'Profile "{profile_name}" not in settings.yaml.')

            return {
                "parameters": self.profiles[profile_name].parameters,
                "credentials": {
                    cred_project_name: self.credentials[cred_settings_name]
                    for cred_project_name, cred_settings_name in self.profiles[
                        profile_name
                    ].credentials.items()
                },
            }

    yaml: Optional[SettingsYaml]
    environment: Optional[Environment]

    def get_settings(self, profile_name=None):
        if profile_name is not None and self.yaml is None:
            return Err("get_settings", "missing_settings_yaml")
        elif self.yaml is not None:
            out = self.yaml.get_profile_info(profile_name)
        else:
            out = {"credentials": dict(), "parameters": dict()}

        if profile_name is None and self.environment is not None:
            # When no profile is specified, and there's something in the environment,
            # we try to use environment variables
            if self.environment.parameters is not None:
                out["parameters"].update(self.environment.parameters)
            if self.environment.credentials is not None:
                out["credentials"].update(self.environment.credentials)

        return Ok(out)


def read_settings():
    environment = {"parameters": dict(), "credentials": dict()}
    for name, value in os.environ.items():
        name = RE_ENV_VAR_NAME.match(name)
        if name is not None:
            name = name.groupdict()

            environment[name["type"].lower() + "s"][name["name"]] = YAML().load(value)

    environment = {k: v for k, v in environment.items() if len(v) > 0}
    if len(environment) == 0:
        environment = None

    filepath = Path("settings.yaml")
    if filepath.exists():
        result = read_yaml_file(filepath)
        if result.is_err:
            return result
        else:
            settings_yaml = result.value
    else:
        settings_yaml = None

    try:
        if settings_yaml is not None:
            return Ok(Settings(yaml=settings_yaml, environment=environment))
        else:
            return Ok(Settings(environment=environment))
    except ValidationError as e:
        return Exc(e, where="settings_reading")


###############################
# Connections functions
###############################


def get_connections(credentials):
    try:
        return Ok(
            {
                name: create_db(name, name, deepcopy(config))
                if config["type"] != "api"
                else {k: v for k, v in config.items() if k != "type"}
                for name, config in credentials.items()
            }
        )
    except Exception as e:
        return Exc(e)


###############################
# Task related config functions
###############################


def get_presets(global_presets, groups):
    """Returns a dictionary of presets merged with the referenced preset

    Presets define a direct acyclic graph by including the `preset` property, so
    this function validates that there are no cycles and that all referenced presets
    are defined.

    In the output, preset names are prefixed with `sayn_global:` or `group:` so that we can
    merge all presets in the project in the same dictionary.

    Args:
      global_presets (dict): dictionary containing the presets defined in project.yaml
      groups (sayn.app.config.TaskGroup): a list of task groups from the tasks/ folder
    """
    # 1. Construct a dictionary of presets so we can attach that info to the tasks
    presets_info = {
        f"sayn_global:{k}": {kk: vv for kk, vv in v.items() if kk != "preset"}
        for k, v in global_presets.items()
    }

    # 1.1. We start with the global presets defined in project.yaml
    presets_project = {
        k: [f"sayn_global:{v}"] if v is not None else []
        for k, v in {
            f"sayn_global:{name}": preset.get("preset")
            for name, preset in global_presets.items()
        }.items()
    }

    # 1.2. Then we add the presets defined in the task groups
    for group_name, group in groups.items():
        presets_info.update(
            {
                f"{group_name}:{k}": {kk: vv for kk, vv in v.items() if kk != "preset"}
                for k, v in group.presets.items()
            }
        )

        group_presets = {
            name: preset.get("preset") for name, preset in group.presets.items()
        }

        # Check if the preset referenced is defined in the task group, otherwise, point at the
        # global task group
        group_presets = {
            f"{group_name}:{k}": [
                f"{group_name}:{v}"
                if v in group_presets and v != k
                else f"sayn_global:{v}"
            ]
            if v is not None
            else []
            for k, v in group_presets.items()
        }
        presets_project.update(group_presets)

    # 1.3. The preset references represent a dag that we need to validate, ensuring
    #      there are no cycles and that all references exists
    result = topological_sort(presets_project)
    if result.is_err:
        return result
    else:
        topo_sort = result.value

    # 1.4. Merge the presets with the reference preset, so that we have 1 dictionary
    #      per preset a task could reference
    presets = {
        name: merge_dict_list(
            [presets_info[p] for p in upstream(presets_project, name).value]
            + [presets_info[name]]
        )
        for name in topo_sort
    }

    return Ok(presets)


def get_task_dict(task, task_name, group_name, presets):
    """Returns a single task merged with the referenced preset

    Args:
      task (dict): a dictionary with the task information
      task_name (str): the name of the task
      group_name (str): the name of the group it appeared on
      presets (dict): a dictionary of merged presets returned by get_presets
    """
    if "preset" in task:
        preset_name = task["preset"]
        preset = presets.get(
            f"{group_name}:{preset_name}",
            presets.get(f"sayn_global:{preset_name}"),
        )
        if preset is None:
            return Err(
                "get_task_dict",
                "missing_preset",
                group=group_name,
                task=task_name,
                preset=preset_name,
            )
        task = merge_dicts(preset, task)

    return Ok(dict(task, name=task_name, group=group_name))


def get_tasks_dict(global_presets, groups):
    """Returns a dictionary with the task definition with the preset information merged

    Args:
      global_presets (dict): a dictionary with the presets as defined in project.yaml
      groups (sayn.common.config.TaskGroup): a list of task groups from the tasks/ folder
    """
    result = get_presets(global_presets, groups)
    if result.is_err:
        return result
    else:
        presets = result.value

    errors = dict()
    tasks = dict()
    for group_name, group in groups.items():
        if group.tasks is not None:
            for task_name, task in group.tasks.items():
                if task_name in tasks:
                    return Err(
                        "dag",
                        "duplicate_task",
                        task=task_name,
                        groups=(group_name, tasks[task_name]["group"]),
                    )
                result = get_task_dict(task, task_name, group_name, presets)
                if result.is_ok:
                    tasks[task_name] = result.value
                else:
                    errors[task_name] = result.error

    if len(errors) > 0:
        return Err("get_tasks_dict", "task_parsing_error", errors=errors)
    else:
        return Ok(tasks)



import random
from uuid import uuid4

from sayn import PythonTask


class LoadData(PythonTask):
    # Dimensions
    fighters = ["Son Goku", "John", "Lucy", "Dr. x", "Carlos", "Dr. y?"]
    arenas = ["Earth Canyon", "World Edge", "Namek", "Volcanic Crater", "Underwater"]
    tournaments = ["World Championships", "Tenka-ichi Budokai", "King of the Mountain"]

    def run(self):
        # Define the run steps
        self.set_run_steps(
            [
                "Generate Dimensions",
                "Generate Battles",
                "Load fighters",
                "Load arenas",
                "Load tournaments",
                "Load battles",
            ]
        )

        with self.step("Generate Dimensions"):
            # Add ids to the dimensions
            fighters = [
                {"fighter_id": str(uuid4()), "fighter_name": val}
                for id, val in enumerate(self.fighters)
            ]
            arenas = [
                {"arena_id": str(uuid4()), "arena_name": val}
                for id, val in enumerate(self.arenas)
            ]
            tournaments = [
                {"tournament_id": str(uuid4()), "tournament_name": val}
                for id, val in enumerate(self.tournaments)
            ]

        with self.step("Generate Battles"):
            battles = list()
            for tournament in tournaments:
                tournament_id = tournament["tournament_id"]
                # Randomly select a number of battles to generate for each tournament
                n_battles = random.choice([10, 20, 30])

                for _ in range(n_battles):
                    battle_id = str(uuid4())

                    # Randomly choose fighters and arena
                    fighter1_id = random.choice(fighters)["fighter_id"]
                    fighter2_id = random.choice(
                        [f for f in fighters if f["fighter_id"] != fighter1_id]
                    )["fighter_id"]
                    arena_id = random.choice(arenas)["arena_id"]

                    # Pick a winner
                    winner_id = (
                        fighter1_id if random.uniform(0, 1) <= 0.5 else fighter2_id
                    )

                    battles.append(
                        {
                            "event_id": str(uuid4()),
                            "tournament_id": tournament_id,
                            "battle_id": battle_id,
                            "arena_id": arena_id,
                            "fighter1_id": fighter1_id,
                            "fighter2_id": fighter2_id,
                            "winner_id": winner_id,
                        }
                    )

        data_to_load = {
            "fighters": fighters,
            "arenas": arenas,
            "tournaments": tournaments,
            "battles": battles,
        }

        # Load logs
        for log_type, log_data in data_to_load.items():
            with self.step(f"Load {log_type}"):
                self.default_db.load_data(f"logs_{log_type}", log_data, replace=True)

        return self.success()



from copy import deepcopy
import os
from typing import List, Dict, Any

from jinja2 import Environment, FileSystemLoader, StrictUndefined

from ..core.errors import Err, Exc, Ok, Result
from ..utils.misc import map_nested

from . import Task, TaskStatus
from .dummy import DummyTask
from .sql import SqlTask
from .autosql import AutoSqlTask
from .copy import CopyTask

_creators = {
    "dummy": DummyTask,
    "sql": SqlTask,
    "autosql": AutoSqlTask,
    "copy": CopyTask,
}

_excluded_properties = (
    "name",
    "type",
    "tags",
    "group",
    "parents",
    "parameters",
    "class",
    "preset",
)


class TaskWrapper:
    """Task wrapper managing the execution of tasks.

    This class wraps a task runner (sayn.tasks.Task) and manages the lifetime of said task.
    In a SAYN execution there will be 1 TaskWrapper per task in the project, whether the task
    is to be executed or ignored based on the task query (`-t` and `-x` arguments to `sayn`).

    Properties:
      name (str): the name of the task
      group (str): the name of the task group file where the task was defined
      tags (List[str]): list of tags declared for the task
      parents (List[sayn.task_wrapper.TaskWrapper]): the list of parents to the current task
      project_parameters (Dict[str, Any]): parameters defined at project level
      task_parameters (Dict[str, Any]): parameters defined at task level
      in_query (bool): whether the task is selected for execution based on the task query
      runner (Task): the object that will do the actual work
      status (TaskStatus): the current status of the task

    """

    name: str = None
    group: str = None
    tags: List[str] = list()
    parents: List[Any] = list()
    project_parameters: Dict[str, Any] = dict()
    task_parameters: Dict[str, Any] = dict()
    in_query: bool = True
    runner: Task = None
    status: TaskStatus = TaskStatus.UNKNOWN

    def check_skip(self):
        failed_parents = {
            p.name: p.status
            for p in self.parents
            if p.status in (TaskStatus.SKIPPED, TaskStatus.FAILED)
        }

        if len(failed_parents) > 0:
            self.status = TaskStatus.SKIPPED

            if self.in_query:
                return Err("task", "parent_errors", failed_parents=failed_parents)
            else:
                return Ok(self.status)

        else:
            return Ok(self.status)

    def setup(
        self,
        task_info,
        parents,
        in_query,
        tracker,
        connections,
        default_db,
        project_parameters,
        run_arguments,
        python_loader,
    ):
        def failed(result):
            self.status = TaskStatus.FAILED
            self.tracker.current_step = None
            return result

        def setup_runner(runner, runner_config):
            try:
                result = runner.setup(**runner_config)

            except Exception as exc:
                result = Exc(exc)

            finally:
                if not isinstance(result, Result):
                    return failed(
                        Err("task_result", "missing_result_error", result=result)
                    )
                elif result.is_err:
                    return failed(result)
                else:
                    self.runner = runner
                    self.status = TaskStatus.READY
                    return Ok()

        self.tracker = tracker

        self.status = TaskStatus.SETTING_UP
        self._info = task_info

        self.name = task_info["name"]
        self.group = task_info["group"]

        self._type = task_info.get("type")
        self.tags = task_info.get("tags", list())
        self.parents = parents

        self.task_parameters = task_info.get("parameters", dict())

        self.in_query = in_query

        # Check the parents are in a good state
        result = self.check_skip()
        if result.is_err or result.value == TaskStatus.SKIPPED:
            return result

        if not in_query:
            self.status = TaskStatus.NOT_IN_QUERY
            return Ok()
        else:

            # Instantiate the Task runner object

            # First get the class to use
            if self._type == "python":
                result = python_loader.get_class("python_tasks", task_info.get("class"))
                if result.is_err:
                    # TODO should probably add the context here
                    return failed(result)
                else:
                    task_class = result.value

            elif self._type in _creators:
                task_class = _creators[self._type]

            else:
                return failed(
                    Err(
                        "task_type",
                        "invalid_task_type_error",
                        type=self._type,
                        group=self.group,
                    )
                )

            # Call the object creation method
            result = self.create_runner(
                task_class,
                {k: v for k, v in task_info.items() if k not in _excluded_properties},
                run_arguments,
                project_parameters,
                default_db,
                connections,
            )
            if result.is_ok:
                runner, runner_config = result.value
            else:
                return failed(result)

            # Run the setup stage for the runner and return the results
            return setup_runner(runner, runner_config)

    def create_runner(
        self,
        task_class,
        runner_config,
        run_arguments,
        project_parameters,
        default_db,
        connections,
    ):
        runner = task_class()

        # Add the basic properties
        runner.name = self.name
        runner.group = self.group
        runner.tags = self.tags
        runner.run_arguments = run_arguments
        env_arguments = {
            "full_load": run_arguments["full_load"],
            "start_dt": f"'{run_arguments['start_dt'].strftime('%Y-%m-%d')}'",
            "end_dt": f"'{run_arguments['end_dt'].strftime('%Y-%m-%d')}'",
        }

        # Process parameters
        # The project parameters go as they come
        runner.project_parameters = deepcopy(project_parameters or dict())

        # Create a jinja environment with the project parameters so that we
        # can use that to compile parameters and other properties
        jinja_env = Environment(
            loader=FileSystemLoader(os.getcwd()),
            undefined=StrictUndefined,
            keep_trailing_newline=True,
        )
        jinja_env.globals.update(task=self)
        jinja_env.globals.update(**env_arguments)
        jinja_env.globals.update(**runner.project_parameters)
        # Compile nested dictionary of parameters
        try:
            runner.task_parameters = map_nested(
                self.task_parameters,
                lambda x: jinja_env.from_string(x).render()
                if isinstance(x, str)
                else x,
            )
        except Exception as e:
            return Exc(e, where="compile_task_parameters")

        # Add the task paramters to the jinja environment
        jinja_env.globals.update(**runner.task_parameters)
        runner.jinja_env = jinja_env

        # Now we can compile the other properties
        try:
            runner_config = map_nested(
                runner_config,
                lambda x: runner.jinja_env.from_string(x).render()
                if isinstance(x, str)
                else x,
            )
        except Exception as e:
            return Exc(e, where="compile_task_properties")

        # Connections and logging
        runner._default_db = default_db
        runner.connections = connections
        runner.tracker = self.tracker

        return Ok((runner, runner_config))

    def should_run(self):
        return self.status == TaskStatus.NOT_IN_QUERY

    def run(self):
        return self.execute_task("run")

    def compile(self):
        return self.execute_task("compile")

    def execute_task(self, command):
        result = self.check_skip()
        if result.is_err or result.value == TaskStatus.SKIPPED:
            return result

        if not self.in_query:
            # TODO review this as it should never happend if running sayn cli
            self.status = TaskStatus.NOT_IN_QUERY
            return Err("execution", "task_not_in_query")
        elif self.status not in (TaskStatus.SETTING_UP, TaskStatus.READY):
            return Err("execution", "setup_error", status=self.status)
        else:
            try:
                if command == "run":
                    result = self.runner.run()
                else:
                    result = self.runner.compile()
                if result.is_ok:
                    self.status = TaskStatus.SUCCEEDED
                else:
                    self.status = TaskStatus.FAILED
            except Exception as e:
                self.status = TaskStatus.FAILED
                result = Exc(e)

            return result



from contextlib import contextmanager
from enum import Enum
from pathlib import Path

from jinja2 import Template

from ..core.errors import Err, Exc, Ok
from ..database import Database


class TaskStatus(Enum):
    NOT_IN_QUERY = "not_in_query"

    SETTING_UP = "setting_up"
    READY = "ready"
    SETUP_FAILED = "setup_failed"

    EXECUTING = "executing"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    SKIPPED = "skipped"

    UNKNOWN = "unknown"


class Task:
    """
    Base class for tasks in SAYN.

    Attributes:
        name (str): Name of the task as defined in the task group.
        group (str): Name of the task group where the task was defined.
        run_arguments (dict): Dictionary containing the values for the arguments specified in the cli.
        task_parameters (dict): Provides access to the parameters specified in the task.
        project_parameters (dict): Provides access to the global parameters of the project.
        parameters (dict): Convinience property joining project and task parameters.
        connections (dict): Dictionary of connections specified for the project.
        tracker (sayn.logging.TaskEventTracker): Message tracker for the current task.
        jinja_env (jinja2.Environment): Jinja environment for this task. The environment comes pre-populated with the parameter values relevant to the task.
    """

    name = None
    group = None
    tags = list()
    run_arguments = dict()
    task_parameters = dict()
    project_parameters = dict()

    _default_db = None
    connections = dict()
    tracker = None

    jinja_env = None

    # Handy properties
    @property
    def parameters(self):
        return {**self.project_parameters, **self.task_parameters}

    @property
    def default_db(self):
        return self.connections[self._default_db]

    # Making it backwards compatible
    @property
    def logger(self):
        return self.tracker

    # Task lifetime methods

    # def setup(self):
    #     raise NotImplementedError("SAYN task", self.__class__.__name__, "setup")

    # def run(self):
    #     raise NotImplementedError("SAYN task", self.__class__.__name__, "run")

    # def compile(self):
    #     raise NotImplementedError("SAYN task", self.__class__.__name__, "compile")

    # Status methods

    def ready(self):
        """(Deprecated: use `success` instead) Returned on successful execution."""
        return Ok()

    def success(self):
        """Returned on successful execution."""
        return Ok()

    def fail(self, msg=None):
        """Returned on failure in any stage."""
        if msg is None:
            msg = 'Unknown error. Use `self.fail("Error message")` in python tasks for more details.'
        return Err("tasks", "task_fail", message=msg)

    # Logging methods

    def set_run_steps(self, steps):
        """Sets the run steps for the task, allowing the CLI to indicate task execution progress. """
        self.tracker.set_run_steps(steps)

    def add_run_steps(self, steps):
        """Adds new steps to the list of run steps for the task, allowing the CLI to indicate task execution progress. """
        self.tracker.add_run_steps(steps)

    def start_step(self, step):
        """Specifies the start of a task step"""
        self.tracker.start_step(step)

    def finish_current_step(self, result=Ok()):
        """Specifies the end of the current step"""
        self.tracker.finish_current_step(result)

    def debug(self, message, details=None):
        """Print a debug message when executing sayn in debug mode (`sayn run -d`)"""
        if details is None:
            details = dict()
        self.tracker.debug(message, **details)

    def info(self, message, details=None):
        """Prints an info message."""
        if details is None:
            details = dict()
        self.tracker.info(message, **details)

    def warning(self, message, details=None):
        """Prints a warning message which will be persisted on the screen after the task concludes execution."""
        if details is None:
            details = dict()
        self.tracker.warning(message, **details)

    def error(self, message, details=None):
        """Prints an error message which will be persisted on the screen after the task concludes execution.

        Executing this method doesn't abort the task or changes the task status. Use `return self.fail` for that instead.

        Args:
          message (str): An optinal error message to print to the screen.

        """
        if details is None:
            details = dict()
        self.tracker.error(message, **details)

    @contextmanager
    def step(self, step):
        """Step context

        Usage:
        ```python
        with self.step('Generate Data'):
            data = generate_data()
        ```

        Args:
          step (str): name of the step being executed.
        """
        self.tracker.start_step(step)
        yield
        self.tracker.finish_current_step()

    # Jinja methods

    def get_template(self, obj):
        """Returns a Jinja template object.

        Args:
          obj (str/Path): The object to transform into a template. If a `pathlib.Path` is specified, the template will be read from disk.

        """
        if isinstance(obj, Path):
            try:
                template = obj.read_text()
            except Exception as e:
                return Err("tasks", "get_template_error", file_path=obj, exception=e)
        elif isinstance(obj, str):
            template = obj
        else:
            return Err("tasks", "get_template_error", obj=obj)

        try:
            return Ok(self.jinja_env.from_string(template))
        except Exception as e:
            return Exc(e, template=template)

    def compile_obj(self, obj, **params):
        """Compiles the object into a string using the task jinja environment.

        Args:
          obj (str/Path/Template): The object to compile. If the object is not a Jinja template object, `self.get_template` will be called first.
          params (dict): An optional dictionary of additional values to use for compilation.
              Note: Project and task parameters values are already set in the environment, so there's no need to pass them on
        """
        if isinstance(obj, Template):
            template = obj
        else:
            result = self.get_template(obj)
            if result.is_err:
                return result
            else:
                template = result.value

        try:
            return Ok(template.render(**params))
        except Exception as e:
            return Exc(e, template=template, params=params)

    def write_compilation_output(self, content, suffix=None):
        path = Path(
            self.run_arguments["folders"]["compile"],
            self.group,
            Path(f"{self.name}{'_'+suffix if suffix is not None else ''}.sql"),
        )

        # Ensure the path exists and it's empty
        path.parent.mkdir(parents=True, exist_ok=True)
        if path.exists():
            path.unlink()

        path.write_text(str(content))

        return Ok()

    # Execution utilities

    def use_db_object(
        self, object_name, schema=None, tmp_schema=None, db=None, request_tmp=True
    ):
        if db is None:
            target_db = self.default_db
        elif isinstance(db, str):
            target_db = self.connections[db]
        elif isinstance(db, Database):
            target_db = db
        else:
            return Err("use_db_object", "wrong_db_type")

        target_db._request_object(
            object_name,
            schema=schema,
            tmp_schema=tmp_schema,
            task_name=self.name,
            request_tmp=request_tmp,
        )


class PythonTask(Task):
    def setup(self):
        self.debug("Nothing to be done")
        return self.success()

    def run(self):
        self.debug("Nothing to be done")
        return self.success()

    def compile(self):
        self.debug("Nothing to be done")
        return self.success()



from typing import Any, Dict, Optional

from pydantic import BaseModel, Field, validator
from sqlalchemy import or_, select, column

from ..core.errors import Err, Exc, Ok
from ..database import Database
from .sql import SqlTask


class Source(BaseModel):
    supports_schemas: bool
    db_type: str

    db_schema: Optional[str] = Field(None, alias="schema")
    table: str
    db: str

    @validator("db_schema")
    def can_use_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'schema not supported for database of type {values["db_type"]}'
            )

        return v


class Destination(BaseModel):
    supports_schemas: bool
    db_type: str

    tmp_schema: Optional[str]
    db_schema: Optional[str] = Field(None, alias="schema")
    table: str
    db: Optional[str]

    @validator("tmp_schema")
    def can_use_tmp_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'tmp_schema not supported for database of type {values["db_type"]}'
            )

        return v

    @validator("db_schema")
    def can_use_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'schema not supported for database of type {values["db_type"]}'
            )

        return v


class Config(BaseModel):
    source: Source
    destination: Destination
    ddl: Optional[Dict[str, Any]]
    delete_key: Optional[str]
    incremental_key: Optional[str]

    @validator("incremental_key", always=True)
    def incremental_validation(cls, v, values):
        if (v is None) != (values.get("delete_key") is None):
            raise ValueError(
                'Incremental copy requires both "delete_key" and "incremental_key"'
            )

        return v


class CopyTask(SqlTask):
    def setup(self, **config):
        conn_names_list = [
            n for n, c in self.connections.items() if isinstance(c, Database)
        ]

        # check the source db exists in settings
        if (
            isinstance(config.get("source"), dict)
            and config["source"].get("db") is not None
        ):
            if config["source"]["db"] not in conn_names_list:
                return Err(
                    "task_definition",
                    "source_db_not_in_settings",
                    db=config["source"]["db"],
                )

        # set the target db for execution
        # this check needs to happen here so we can pass db_features and db_type to the validator
        if (
            isinstance(config.get("destination"), dict)
            and config["destination"].get("db") is not None
        ):
            if config["destination"]["db"] not in conn_names_list:
                return Err(
                    "task_definition",
                    "destination_db_not_in_settings",
                    db=config["destination"]["db"],
                )
            self._target_db = config["destination"]["db"]
        else:
            self._target_db = self._default_db

        if isinstance(config.get("source"), dict):
            config["source"].update(
                {
                    "supports_schemas": not self.connections[
                        config["source"]["db"]
                    ].feature("NO SCHEMA SUPPORT"),
                    "cannot_change_schema": self.connections[
                        config["source"]["db"]
                    ].feature("CANNOT CHANGE SCHEMA"),
                    "db_type": self.connections[config["source"]["db"]].db_type,
                }
            )

        if isinstance(config.get("destination"), dict):
            config["destination"].update(
                {
                    "supports_schemas": not self.target_db.feature("NO SCHEMA SUPPORT"),
                    "cannot_change_schema": self.target_db.feature(
                        "CANNOT CHANGE SCHEMA"
                    ),
                    "db_type": self.target_db.db_type,
                }
            )

        try:
            self.config = Config(**config)
        except Exception as e:
            return Exc(e)

        self.source_db = self.connections[self.config.source.db]
        self.source_schema = self.config.source.db_schema
        self.source_table = self.config.source.table
        self.use_db_object(
            self.source_table,
            schema=self.source_schema,
            db=self.source_db,
            request_tmp=False,
        )

        self.tmp_schema = (
            self.config.destination.tmp_schema or self.config.destination.db_schema
        )
        self.schema = self.config.destination.db_schema
        self.table = self.config.destination.table
        self.tmp_table = f"sayn_tmp_{self.table}"

        self.use_db_object(
            self.table,
            schema=self.schema,
            tmp_schema=self.tmp_schema,
            db=self.target_db,
            request_tmp=True,
        )

        self.delete_key = self.config.delete_key
        self.incremental_key = self.config.incremental_key

        self.is_full_load = self.run_arguments["full_load"] or self.delete_key is None

        result = self.target_db._validate_ddl(self.config.ddl)
        if result.is_ok:
            self.ddl = result.value
        else:
            return result

        return Ok()

    def compile(self):
        return self.execute(False, self.run_arguments["debug"])

    def run(self):
        return self.execute(True, self.run_arguments["debug"])

    def execute(self, execute, debug):
        # Introspect target
        self.target_table_exists = self.target_db._table_exists(self.table, self.schema)

        steps = ["Prepare Load", "Load Data"]
        if self.target_table_exists:
            load_table = self.tmp_table
            load_schema = self.tmp_schema
            if self.is_full_load or self.incremental_key is None:
                steps.append("Move Table")
            else:
                steps.append("Merge Tables")
        else:
            load_table = self.table
            load_schema = self.schema

        self.set_run_steps(steps)

        with self.step("Prepare Load"):
            result = self.get_columns()
            if result.is_err:
                return result

            result = self.get_read_query(execute, debug)
            if result.is_err:
                return result
            else:
                get_data_query = result.value

            query = self.target_db.create_table(
                load_table, schema=load_schema, replace=True, **self.ddl
            )
            if debug:
                self.write_compilation_output(query, "create_table")
            if execute:
                try:
                    self.target_db.execute(query)
                except Exception as e:
                    return Exc(e)

        with self.step("Load Data"):
            if execute:
                data_iter = self.source_db._read_data_stream(get_data_query)
                self.target_db.load_data(load_table, data_iter, schema=load_schema)

        # Final step
        final_step = steps[-1]
        if final_step == "Move Table":
            query = self.target_db.move_table(
                load_table,
                self.table,
                src_schema=load_schema,
                dst_schema=self.schema,
                **self.ddl,
            )
        elif final_step == "Merge Tables":
            query = self.target_db.merge_tables(
                load_table,
                self.table,
                self.delete_key,
                src_schema=load_schema,
                dst_schema=self.schema,
                **self.ddl,
            )
        else:
            query = None

        if query is not None:
            with self.step(final_step):
                if debug:
                    self.write_compilation_output(
                        query, final_step.replace(" ", "_").lower()
                    )
                if execute:
                    try:
                        self.target_db.execute(query)
                    except Exception as e:
                        return Exc(e)

        return Ok()

    def get_columns(self):
        # We get the source table definition
        source_table_def = self.source_db._get_table(
            self.source_table,
            self.source_schema,
            # columns=[c["name"] for c in self.ddl["columns"]],
            # required_existing=True,
        )
        if source_table_def is None:
            return Err(
                "database_error",
                "source_db_missing_source_table",
                schema=self.source_schema,
                table=self.source_table,
                db=self.source_db.name,
            )
        self.source_table_def = source_table_def

        if len(self.ddl["columns"]) == 0:
            dst_table_def = None
            if not self.is_full_load:
                dst_table_def = self.target_db._get_table(self.table, self.schema)

            if dst_table_def is not None:
                # In incremental loads we use the destination table to determine the columns
                self.ddl["columns"] = [
                    {
                        "name": c.name,
                        "type": c.type.compile(dialect=self.target_db.engine.dialect),
                    }
                    for c in dst_table_def.columns
                ]

                # Ensure these columns are in the source
                missing_columns = set([c.name for c in dst_table_def.columns]) - set(
                    [c.name for c in self.source_table_def.columns]
                )
                if len(missing_columns) > 0:
                    return Err(
                        "database_error",
                        "source_table_missing_columns",
                        db=self.source_db.name,
                        table=self.source_table,
                        schema=self.source_schema,
                        columns=missing_columns,
                    )

            else:
                # In any other case, we use the source
                self.ddl["columns"] = [
                    {"name": c.name, "type": self.target_db._py2sqa(c.type.python_type)}
                    for c in self.source_table_def.columns
                ]
        else:
            # Fill up column types from the source table
            for col in self.ddl["columns"]:
                if col.get("name") not in self.source_table_def.columns:
                    return Err(
                        "database_error",
                        "source_table_missing_column",
                        db=self.source_db.name,
                        table=self.source_table,
                        schema=self.source_schema,
                        column=col.get("name"),
                    )

                if "type" not in col:
                    col["type"] = self.target_db._py2sqa(
                        self.source_table_def.columns[col["name"]].type.python_type
                    )

        return Ok()

    def get_read_query(self, execute, debug):
        # Get the incremental value
        last_incremental_value_query = (
            f"SELECT MAX({self.incremental_key}) AS value\n"
            f"FROM {'' if self.schema is None else self.schema +'.'}{self.table}\n"
            f"WHERE {self.incremental_key} IS NOT NULL"
        )
        if debug:
            self.write_compilation_output(
                last_incremental_value_query, "last_incremental_value"
            )

        get_data_query = select(
            columns=[column(c["name"]) for c in self.ddl["columns"]],
            from_obj=self.source_table_def,
        )
        last_incremental_value = None

        if not self.is_full_load and self.target_table_exists:
            if execute:
                res = self.target_db.read_data(last_incremental_value_query)
                if len(res) == 1:
                    last_incremental_value = res[0]["value"]
            else:
                last_incremental_value = "LAST_INCREMENTAL_VALUE"

        # Select stream
        if last_incremental_value is not None:
            get_data_query = get_data_query.where(
                or_(
                    self.source_table_def.c[self.incremental_key].is_(None),
                    self.source_table_def.c[self.incremental_key]
                    > last_incremental_value,
                )
            )
        if debug:
            self.write_compilation_output(get_data_query, "get_data")

        return Ok(get_data_query)



from sqlalchemy import create_engine

from . import Database

db_parameters = [
    "account",
    "region",
    "user",
    "password",
    "database",
    "warehouse",
    "role",
    "schema",
]


class Snowflake(Database):
    def feature(self, feature):
        return feature in ("TABLE RENAME CHANGES SCHEMA",)

    def create_engine(self, settings):
        from snowflake.sqlalchemy import URL

        url_params = dict()
        for param in db_parameters:
            if param in settings:
                url_params[param] = settings.pop(param)

        return create_engine(URL(**url_params), **settings)

    def execute(self, script):
        conn = self.engine.connect()
        conn.connection.execute_string(script)
        conn.connection.commit()
        conn.connection.close()
